/*
    Copyright 2008 Chris Evans
    cevans42ca at yahoo.ca

    This file is part of the SyncDocs project, available at:
    http://syncdocs.sourceforge.net/

    SyncDocs is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    SyncDocs is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SyncDocs; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package ca.quine.syncdocs;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.*;

import javax.swing.*;
import javax.swing.table.DefaultTableModel;
import javax.swing.table.TableColumn;

/**
 * Posted by Aberghouss; modified by Chris Evans.
 * 
 * @see http://forums.sun.com/thread.jspa?threadID=5245523&messageID=10012580
 */
public class DropPanel extends JPanel implements FileReceivedHandler, PropertyChangeListener {

	private static final long serialVersionUID = 7271333854200223911L;

	public static final int INITIAL_WIDTH = 600;
	public static final int INITIAL_HEIGHT = 200;

	private static final int FILE_COLUMN = 0;
	private static final int APP_NAME_COLUMN = 1;

	private String displaySyncWithNext;
	private String displayDontSync;

	private JScrollPane scrollPane;
	private DefaultTableModel tableModel;
	private JTable fileList;

	private Messages messages;
	private MyTransferHandler transferHandler;

	public DropPanel(SyncDocs syncDocs) {
		messages = syncDocs.getMessages();
		messages.registerListener(this);

		transferHandler = new MyTransferHandler();

		displaySyncWithNext = syncDocs.getLocalisedString("DropPanel.label.displaySyncWithNext"); //$NON-NLS-1$
		displayDontSync = syncDocs.getLocalisedString("DropPanel.label.displayDontSync"); //$NON-NLS-1$

		tableModel = new SyncDocsTableModel();
		Object[] columnHeaders = new Object[2];
		columnHeaders[0] = syncDocs.getLocalisedString("DropPanel.column.fileName"); //$NON-NLS-1$
		columnHeaders[1] = syncDocs.getLocalisedString("DropPanel.column.appName"); //$NON-NLS-1$
		tableModel.setColumnIdentifiers(columnHeaders);

		addFileReceivedHandler(this);
		setTransferHandler(transferHandler);

		fileList = new JTable(tableModel);
		fileList.setTransferHandler(transferHandler);
		fileList.setSelectionMode(ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);
		fileList.setIntercellSpacing(new Dimension(2, 2));

		TableColumn fileColumn = fileList.getColumn(columnHeaders[0]);
		fileColumn.setPreferredWidth(430);

		TableColumn appColumn = fileList.getColumn(columnHeaders[1]);
		appColumn.setPreferredWidth(170);

		JComboBox comboBox = new JComboBox();
		comboBox.setEditable(false);
		comboBox.addItem(displaySyncWithNext);
		comboBox.addItem(displayDontSync);
		comboBox.setSelectedIndex(1);

		appColumn.setCellEditor(new DefaultCellEditor(comboBox));

		setBorder(BorderFactory.createBevelBorder(1));

		scrollPane = new JScrollPane(fileList);
		scrollPane.setTransferHandler(transferHandler);
		scrollPane.setMinimumSize(new Dimension(INITIAL_WIDTH, INITIAL_HEIGHT));
		scrollPane.setPreferredSize(new Dimension(INITIAL_WIDTH, INITIAL_HEIGHT));

		BorderLayout layout = new BorderLayout();
		setLayout(layout);

		add(scrollPane);
	}

	public void removeSelected() {
		int[] selectedIndices = fileList.getSelectedRows();
		if (selectedIndices.length == 0) {
			return;
		}

		do {
			// All the indices will change for the selected items every time we
			// delete an item.
			int index = selectedIndices[0];
			tableModel.removeRow(index);

			selectedIndices = fileList.getSelectedRows();
		}
		while (selectedIndices.length > 0);
	}

	public void addFileReceivedHandler(FileReceivedHandler handler) {
		transferHandler.addFileReceivedHandler(handler);
	}

	public void removeFileReceivedHandler(FileReceivedHandler handler) {
		transferHandler.removeFileReceivedHandler(handler);
	}

	public void receiveFileName(String filename) {
		if (contains(tableModel, filename)) {
			JOptionPane.showMessageDialog(null,
					messages.getString("DropPanel.label.filenameAlreadyAdded")); //$NON-NLS-1$

			return;
		}

		Object[] rowObjects = new Object[2];
		rowObjects[0] = filename;
		rowObjects[1] = displaySyncWithNext;

		tableModel.addRow(rowObjects);
	}

	private boolean contains(DefaultTableModel listModel, String filename) {
		int rows = listModel.getRowCount();
	
		for (int row = 0; row < rows; row++) {
			String filenameInTable = (String) listModel.getValueAt(row,
					FILE_COLUMN);
	
			if (filenameInTable.equals(filename)) {
				return true;
			}
		}
	
		return false;
	}

	public List getFilenameList() {
		ArrayList retVal = new ArrayList();

		int rows = tableModel.getRowCount();
		for (int row = 0; row < rows; row++) {
			String filename = (String) tableModel.getValueAt(row, FILE_COLUMN);
			String appId =
					(String) tableModel.getValueAt(row, APP_NAME_COLUMN);

			if (appId.equals(displayDontSync)) {
				appId = FileEntry.DONT_SYNC;
			}
			else if (appId.equals(displaySyncWithNext)) {
				appId = FileEntry.SYNC_WITH_NEXT;
			}

			retVal.add(new FileEntry(filename, appId));
		}

		return retVal;
	}

	public void addFiles(List files) {
		Iterator filenameIter = files.iterator();
		while (filenameIter.hasNext()) {
			FileEntry fileEntry = (FileEntry) filenameIter.next();

			Object[] rowObjects = new Object[2];
			rowObjects[0] = fileEntry.getFileName();

			if (SyncDocs.isEmpty(fileEntry.getAppId())
					|| fileEntry.getAppId().equals(FileEntry.SYNC_WITH_NEXT))
			{
				rowObjects[1] = displaySyncWithNext;
			}
			else if (fileEntry.getAppId().equals(FileEntry.DONT_SYNC)) {
				rowObjects[1] = displayDontSync;
			}
			else {
				rowObjects[1] = fileEntry.getAppId();
			}

			tableModel.addRow(rowObjects);
		}
	}

	public void removeFiles(String appId, String appName, List inputFileList) {
		String displayName = appId;
		if (!SyncDocs.isEmpty(appName)) {
			displayName = appName;
		}

		String message = displayName + " " + messages.getString("DropPanel.label.removeRequestReceived"); //$NON-NLS-1$
		JOptionPane pane = new JOptionPane(message);
        pane.setMessageType(JOptionPane.INFORMATION_MESSAGE);      
        JDialog dialog = pane.createDialog(this, messages.getString("DropPanel.title.removeRequestReceived")); //$NON-NLS-1$
        dialog.setModal(false);
        dialog.setVisible(true);

		Set fileSet = new HashSet(inputFileList);

		int rowCount = tableModel.getRowCount();
		int currentRow = 0;

		while (currentRow < rowCount) {
			String filename = (String) tableModel.getValueAt(currentRow, FILE_COLUMN);
			if (fileSet.contains(filename)) {
				tableModel.removeRow(currentRow);
				rowCount--;
			}
			else {
				currentRow++;
			}
		}
	}

	public void propertyChange(PropertyChangeEvent event) {
		if (event.getSource().equals(messages)) {
			String newDisplaySyncWithNext = messages.getString("DropPanel.label.displaySyncWithNext"); //$NON-NLS-1$
			String newDisplayDontSync = messages.getString("DropPanel.label.displayDontSync"); //$NON-NLS-1$

			Object[] columnHeaders = new Object[2];
			columnHeaders[0] = messages.getString("DropPanel.column.fileName"); //$NON-NLS-1$
			columnHeaders[1] = messages.getString("DropPanel.column.appName"); //$NON-NLS-1$
			tableModel.setColumnIdentifiers(columnHeaders);

			for (int i=0; i<tableModel.getRowCount(); i++) {
				String appName = (String) tableModel.getValueAt(i, APP_NAME_COLUMN);
				if (appName.equals(displayDontSync)) {
					tableModel.setValueAt(newDisplayDontSync, i, APP_NAME_COLUMN);
				}
				else if (appName.equals(displaySyncWithNext)) {
					tableModel.setValueAt(newDisplaySyncWithNext, i, APP_NAME_COLUMN);
				}
			}

			displaySyncWithNext = newDisplaySyncWithNext;
			displayDontSync = newDisplayDontSync;
		}
	}

}
