/*
    Copyright 2008 Chris Evans
    cevans42ca at yahoo.ca

    This file is part of the SyncDocs project, available at:
    http://syncdocs.sourceforge.net/

    SyncDocs is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    SyncDocs is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SyncDocs; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package ca.quine.syncdocs.io;

import java.io.IOException;
import java.io.InputStream;

/**
 * A PUT request from an HTTP client has several lines of headers which
 * are best handled with "readLine()" on a BufferedReader, but then might
 * be followed by a binary stream, which is best handled with an InputStream.
 * The HybridBufferedStreamReader handles switching between the line by line
 * mode and reading a bufferful of data.
 */
public class HybridBufferedStreamReader {

	private static final int BUFFER_SIZE = 4096;
	private byte[] buffer = new byte[BUFFER_SIZE];
	private int bufferStart = 0;
	private int bufferEnd = -1;

	private InputStream is;

	public HybridBufferedStreamReader(InputStream is) {
		this.is = is;
	}

	public String readLine() throws IOException {
		if (bufferEnd > bufferStart) {
			StringBuffer sb = new StringBuffer();

noMore:		while (true) {
				while (bufferStart < bufferEnd) {
					if (buffer[bufferStart] == '\r') {
						bufferStart++;
						continue;
					}

					if (buffer[bufferStart] == '\n') {
						bufferStart++;
						break noMore;
					}

					sb.append((char) buffer[bufferStart]);
					bufferStart++;
				}

				if (!getMore()) {
					break noMore;
				}
			}

			return sb.toString();
		}
		else {
			if (getMore()) {
				return readLine();
			}
			else {
				return null;
			}
		}
	}

	private boolean getMore() throws IOException {
		int length = is.read(buffer);
		if (length == -1) {
			return false;
		}

		bufferStart = 0;
		bufferEnd = length;

		return true;
	}
	public int read(byte[] inputBuffer, int contentLength) throws IOException {
		if (bufferEnd > bufferStart) {
			int length = bufferEnd - bufferStart;
			if (length > inputBuffer.length) {
				length = inputBuffer.length;
			}

			System.arraycopy(buffer, bufferStart, inputBuffer, 0, length);
			bufferStart += length;
			return length;
		}
		else {
			return is.read(inputBuffer, 0, contentLength);
		}
	}

	public void close() throws IOException {
		is.close();
	}

	public int available() throws IOException {
		return is.available();
	}

}
