/*
    Copyright 2008 Chris Evans
    cevans42ca at yahoo.ca
  
    This file is part of the SyncDocs project, available at:
    http://syncdocs.sourceforge.net/

    SyncDocs is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    SyncDocs is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License
    along with SyncDocs; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
    This file is based on TeenyWeb.java by Tom Valesky, available at
    http://patriot.net/~tvalesky//TeenyWeb.html
    Chris Evans's changes to TeenyWeb are licenced under the GPL by
    permission from the original author.

    This version includes some builtin "CGI" like functionality, support
    for PUT requests, MD5 verification and full binary file support.
*/

package ca.quine.web;

import java.io.*;
import java.net.*;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import ca.quine.syncdocs.SyncDocs;
import ca.quine.syncdocs.io.HybridBufferedStreamReader;

public class TeenyWeb extends Thread {

	private ServerSocket listen_socket;
	private IRequestHandler requestHandler;
	private boolean shuttingDown = false;

	public TeenyWeb(int port, IRequestHandler handler) throws IOException {
		// set instance variables from constructor args
		requestHandler = handler;

		// create new ServerSocket
		listen_socket = new ServerSocket(port);

		// Start running Server thread
		this.start();
	}

	public void stopServer() throws IOException {
		shuttingDown = true;
		listen_socket.close();
	}

	public void run() {
		try {
			while (true) {
				// listen for a request. When a request comes in,
				// accept it, then create a Connection object to
				// service the request and go back to listening on
				// the port.

				Socket client_socket = listen_socket.accept();
				new Connection(client_socket, requestHandler);
			}
		}
		catch (SocketException e) {
			if (!shuttingDown) {
				e.printStackTrace();
			}
		}
		catch (IOException e) {
			System.err.println(e);
		}
	}

}

// The Connection class -- this is where HTTP requests are serviced

class Connection extends Thread {
	private static int BUFFER_SIZE = 4096;
	private static Pattern CONTENT_LENGTH_PATTERN = Pattern.compile("^Content-Length: (\\d*)$");

	protected Socket client;

	protected PrintStream out;

	//private String requestedFile;
	private IRequestHandler requestHandler;

	private boolean debug = false;

	public Connection(Socket client_socket, IRequestHandler handler) {
		// set instance variables from args
		client = client_socket;
		requestHandler = handler;

		// create input and output streams for conversation with client

		try {
			out = new PrintStream(client.getOutputStream(), true, "UTF-8");
		}
		catch (IOException e) {
			System.err.println(e);

			try {
				client.close();
			}
			catch (IOException e2) {
				// We could experience the same error again when we try to
				// clean up, in which case it really doesn't matter.
			}

			return;
		}

		// start this object's thread -- the rest of the action
		// takes place in the run() method, which is called by
		// the thread.
		this.start();
	}

	public void run() {
		String line = null; // read buffer
		String req = null; // first line of request
		// OutputStream os;

		try {
			// read HTTP request -- the request comes in
			// on the first line, and is of the form:
			// GET <filename> HTTP/1.x

			HybridBufferedStreamReader br =
					new HybridBufferedStreamReader(client.getInputStream());

			req = br.readLine();
			System.out.print("Request received:  ");
			System.out.println(req);

			// parse request -- get filename
			StringTokenizer st = new StringTokenizer(req);
			String method = st.nextToken();
			String cookedRequest = st.nextToken();

			ParsedHTTPRequest request = ParsedHTTPRequest.parseRequest(cookedRequest);

//			String appNameHeaderValue = null;
//			String appIdHeaderValue = null;

			int contentLength = 0;
			// loop through and discard rest of request
			while ((line = br.readLine()) != null) {
				if (debug) { System.out.println(line); }

				Matcher matcher = CONTENT_LENGTH_PATTERN.matcher(line);
				if (matcher.matches()) {
					contentLength = Integer.parseInt(matcher.group(1));
				}
//				else if (line.startsWith("X-AppName: ")) {
//					appNameHeaderValue = line.substring("X-AppName: ".length());
//				}
//				else if (line.startsWith("X-AppId: ")) {
//					appIdHeaderValue = line.substring("X-AppId: ".length());
//				}

				if (line.trim().length() == 0) {
					break;
				}
			}

			if (method.equals("PUT")) {
				if (contentLength == 0) {
					System.err.println("Client asked to PUT a file but did not indicate how big it was.  Denied.");

					InputStream is = sendBadRequest(400, "Bad Request", "Content-length must be specified.");
					is.close();
					br.close();

					return;
				}

				// If we need to support the HTTP 1.1 based
				// "Expect 100 Continue", the following code is likely what
				// will be required.
				//
				// out.println("HTTP/1.1 100 Continue");
				// out.println("");

				File outputDir = requestHandler.getDownloadsDirectory(request);

				File outputFile = new File(outputDir, request.getFilename());
				String canonicalOutputPath = outputFile.getCanonicalPath();
				String canonicalOutputDir = outputDir.getCanonicalPath();

				// Check for ../ and other craziness.
				if (!canonicalOutputPath.startsWith(canonicalOutputDir)) {
					InputStream is = sendBadRequest(400, "Bad Request", "Invalid directory for PUT.");
					is.close();
					br.close();

					return;
				}

				outputDir.mkdirs();

				if (!outputFile.getParentFile().exists()) {
					if (!outputFile.getParentFile().mkdirs()) {
						InputStream is = sendBadRequest(500, "Internal Server Error", "Unable to create requested directory.");
						is.close();
						br.close();

						return;
					}
				}

				System.out.println("Receiving file as '" + canonicalOutputPath + "'.");
				FileOutputStream fos = new FileOutputStream(outputFile);
				transfer(br, fos, contentLength);
				fos.close();

				try {
					MessageDigest md5 = MessageDigest.getInstance("MD5");
					byte[] inputFileInBytes = getInputFileAsBytes(outputFile);
					byte[] digest = md5.digest(inputFileInBytes);
					String digestString = SyncDocs.convertBytesToHexString(digest);

					Map extraParamMap = new HashMap();
					extraParamMap.put("X-Content-Received-MD5", digestString);
					sendResponseHeader(200, "OK", "text/plain", extraParamMap);
					client.getOutputStream().write(("Received file with digest:  '" + digestString + "'.\r\n").getBytes("ISO-8859-1"));
				}
				catch (NoSuchAlgorithmException e) {
					client.getOutputStream().write("Unable to compute hash:  MD5 not supported.".getBytes("ISO-8859-1"));
				}

				br.close();

				try {
					client.getInputStream().close();
				}
				catch (IOException e) {}

				System.out.println("Received file '" + outputFile.getCanonicalPath() + "' successfully.");

				return;
			}

			if (debug) { System.out.println("Asking request handler for '" + request + "'."); }

			// Check if the request is something we can deal with, if not,
			// send back a 404.
			if (!requestHandler.canRead(request)) {
				sendResponseHeader(404, "File not found", "text/plain");
				sendString("404: not found: " + request);
				out.flush();
				client.close();
				return;
			}

			String contentType = requestHandler.getContentType(request);
			if (contentType == null) {
				contentType = "application/octet-stream";
			}

			// Send 200 OK response.
			sendResponseHeader(200, "OK", contentType);

			// Transfer the file.
			InputStream is = requestHandler.getInputStream(request);
			transfer(is, out);
			out.flush();
			is.close();
			System.out.println("Successfully finished sending file.");
		}
		catch (IOException e) {
			System.out.println(e);
		}
		finally {
			try {
				client.close();
			}
			catch (IOException e) {
				// We could experience the same error again when we try to
				// clean up, in which case it really doesn't matter.
			}
		}
	}

	private InputStream sendBadRequest(int responseCode, String codeText,
			String failDetails) throws UnsupportedEncodingException, IOException
	{
		sendResponseHeader(responseCode, codeText, "text/plain");
		InputStream is = new ByteArrayInputStream(failDetails.getBytes("UTF-8"));
		transfer(is, out);
		out.flush();
		return is;
	}

	private byte[] getInputFileAsBytes(File outputFile)
			throws IOException
	{
		InputStream is = new FileInputStream(outputFile);

		try {
			int fileLength = (int) outputFile.length();
			byte[] wholeFile = new byte[fileLength];
			int bytesRead = is.read(wholeFile);
			if (bytesRead != fileLength) {
				throw new IOException("File is size " + fileLength + ", but only " + bytesRead + " bytes were read.");
			}
	
			return wholeFile;
		}
		finally {
			if (is != null) {
				try {
					is.close();
				}
				catch (IOException e) {
					e.printStackTrace();
				}
			}
		}
	}

	private void transfer(InputStream is, OutputStream os) throws IOException {
		byte[] buffer = new byte[BUFFER_SIZE];

		while (true) {
			int len = is.read(buffer);

			if (len == -1) {
				break;
			}

			// Extra transfer logging:
			//System.out.println("About to write " + len + " bytes from the buffer to the stream.");
			os.write(buffer, 0, len);
			//System.out.println("Finished writing " + len + " bytes from the buffer to the stream.");
		}
	}

	private void transfer(HybridBufferedStreamReader reader,
			OutputStream os, int contentLength) throws IOException
	{
		byte[] buffer = new byte[BUFFER_SIZE];

		int len = -1;
		while (contentLength > 0) {
			if (contentLength < buffer.length) {
				len = reader.read(buffer, contentLength);
			}
			else {
				len = reader.read(buffer, buffer.length);
			}

			if (len == -1) {
				break;
			}

			os.write(buffer, 0, len);
			contentLength -= len;
		}
	}

	void sendResponseHeader(int code, String message, String type) {
		sendResponseHeader(code, message, type, null);
	}

	// send an HTTP header to the client
	// The first line is a status message from the server to the client.
	// The second line holds the mime type of the document
	void sendResponseHeader(int code, String message, String type, Map extraParamMap) {
		out.println("HTTP/1.0 " + code + " " + message);
		out.println("Content-type: " + type);
		if (extraParamMap != null) {
			Iterator paramIter = extraParamMap.entrySet().iterator();
			while (paramIter.hasNext()) {
				Map.Entry entry = (Map.Entry) paramIter.next();
				out.println(entry.getKey() + ": " + entry.getValue());
			}
		}

		out.println("");
	}

	// write a string to the client.
	void sendString(String str) {
		out.print(str);
	}

}
